"""This small module implements generic tests for distributions to ensure that they have been implemented consistently"""

import os
import numpy as np

from qp.ensemble import Ensemble
from qp.plotting import plot_native, plot, plot_pdf_samples_on_axes
from qp.test_data import NPDF
from qp.factory import read, convert


def assert_all_close(arr, arr2, **kwds):
    """A slightly more informative version of asserting allclose"""
    test_name = kwds.pop('test_name', 'test')
    if not np.allclose(arr, arr2, **kwds): #pragma: no cover
        raise ValueError("%s %.2e %.2e %s" % (test_name, (arr-arr2).min(), (arr-arr2).max(), kwds))


def assert_all_small(arr, **kwds):
    """A slightly more informative version of asserting allclose"""
    test_name = kwds.pop('test_name', 'test')
    if not np.allclose(arr, 0, **kwds): #pragma: no cover
        raise ValueError("%s %.2e %.2e %s" % (test_name, arr.min(), arr.max(), kwds))


def build_ensemble(test_data):
    """Build an ensemble from test data in a class"""
    gen_func = test_data['gen_func']
    ctor_data = test_data['ctor_data']
    try:
        ens = Ensemble(gen_func, data=ctor_data)
        ancil = test_data.get('ancil')
        if ancil is not None:
            ens.set_ancil(ancil)
        return ens
    except Exception as exep: #pragma: no cover
        print("Failed to make %s %s %s" % (gen_func, ctor_data, exep))
        raise ValueError from exep


def pdf_func_tests(pdf, test_data, short=False, check_props=True):
    """Run the test for a practicular class"""

    xpts = test_data['test_xvals']

    if check_props:
        # if we used the c'tor, make sure the class keeps the data used in the c'tor
        for kv in test_data['ctor_data'].keys():
            test_val = pdf.kwds.get(kv, None)
            if test_val is None:
                if not hasattr(pdf.dist, kv): #pragma: no cover
                    raise ValueError("%s %s" % (pdf.dist, kv))
                _ = getattr(pdf.dist, kv)


    if hasattr(pdf.dist, 'npdf'):
        assert pdf.dist.npdf == pdf.npdf

    assert pdf.shape[-1] == NPDF

    pdfs = pdf.pdf(xpts)
    if pdf.ndim == 1:
        xslice = xpts[5]
        pdfs_slice = pdf.pdf(xslice)
        check_pdf = pdfs[:,5].flatten() - pdfs_slice.flatten()
        assert_all_small(check_pdf, atol=2e-2, test_name="pdf")

        xslice = np.expand_dims(xpts[np.arange(pdf.npdf)], -1)
        pdfs_slice = pdf.pdf(xslice)
        pdf_check = np.array([pdfs[i,i] for i in range(pdf.npdf)])
        check_pdfs_slice = pdf_check.flatten()  - pdfs_slice.flatten()
        assert_all_small(check_pdfs_slice, atol=2e-2, test_name="pdf_slice")


    if short: #pragma: no cover
        return pdf

    cdfs = pdf.cdf(xpts)
    quants = np.linspace(0.01, 0.99, 50)

    binw = xpts[1:] - xpts[0:-1]
    if pdf.ndim == 1:
        check_cdf = ((pdfs[:,0:-1] + pdfs[:,1:]) * binw /2).cumsum(axis=-1) - cdfs[:,1:]
    else:
        check_cdf = ((pdfs[:,:,0:-1] + pdfs[:,:,1:]) * binw /2).cumsum(axis=-1) - cdfs[:,:,1:]

    assert_all_small(check_cdf, atol=2e-1, test_name="cdf")

    ppfs = pdf.ppf(quants)
    check_ppf = pdf.cdf(ppfs) - quants
    assert_all_small(check_ppf, atol=2e-2, test_name="ppf")

    if pdf.ndim == 1:
        quants_slice = np.expand_dims(quants[np.arange(pdf.npdf)], -1)
        ppfs_slice = pdf.ppf(quants_slice)
        _ = np.array([ppfs[i,i] for i in range(pdf.npdf)])

        check_ppfs_slice = pdf.cdf(ppfs_slice) - quants_slice
        assert_all_small(check_ppfs_slice, atol=2e-2, test_name="ppf_slice")

    sfs = pdf.sf(xpts)
    check_sf = sfs + cdfs
    assert_all_small(check_sf - 1, atol=1e-5, test_name="sf")

    _ = pdf.isf(quants)
    check_isf = pdf.cdf(ppfs) + quants[::-1]
    assert_all_small(check_isf - 1, atol=5e-2, test_name="isf")
    return pdf


def run_pdf_func_tests(test_class, test_data, short=False, check_props=True):
    """Run the test for a practicular class"""

    method = test_data.get('method', None)
    ctor_func = test_class.creation_method(method)
    if ctor_func is None: #pragma: no cover
        raise KeyError("failed to find creation method %s for class %s" % (method, test_class))
    try:
        pdf = ctor_func(**test_data['ctor_data'])
    except Exception as exep: #pragma: no cover
        print("Failed to make %s %s %s" % (ctor_func, test_data['ctor_data'], exep))
        raise ValueError from exep

    alloc_kwds = pdf.dist.get_allocation_kwds(pdf.npdf, **test_data['ctor_data'])
    for key, val in alloc_kwds.items():
        assert np.product(val[0]) == np.size(test_data['ctor_data'][key])

    return pdf_func_tests(pdf, test_data, short=short, check_props=check_props)


def persist_func_test(ensemble, test_data):
    """Run loopback persistence tests on an ensemble"""
    #ftypes = ['fits', 'hdf5', 'pq']
    if ensemble.ndim == 1:
        ftypes = ['fits', 'hf5', 'h5', 'pq', 'hdf5']
    else:
        ftypes = ['fits', 'hf5']

    for ftype in ftypes:
        filename = "test_%s.%s" % ( ensemble.gen_class.name, ftype )
        try:
            os.remove(filename)
        except FileNotFoundError:
            pass
        ensemble.write_to(filename)
        ens_r = read(filename)
        diff = ensemble.pdf(test_data['test_xvals']) - ens_r.pdf(test_data['test_xvals'])
        assert_all_small(diff, atol=1e-5, test_name="persist")
        if ensemble.ancil is not None:
            if ftype in ['pq']:
                continue
            diff2 = ensemble.ancil['zmode'] - ens_r.ancil['zmode']
            assert_all_small(diff2, atol=1e-5, test_name="persist_ancil")
        try:
            os.unlink(filename)
        except FileNotFoundError:
            pass
        try:
            os.unlink(filename.replace(".%s" % ftype, "data.%s" % ftype))
            os.unlink(filename.replace(".%s" % ftype, "meta.%s" % ftype))
        except FileNotFoundError:
            pass


def run_persist_func_tests(test_data):
    """Run the test for a practicular class"""
    ens = build_ensemble(test_data)
    persist_func_test(ens, test_data)


def run_convert_tests(ens_orig, gen_class, test_data, **kwargs):
    """Run the test for a practicular class"""

    xpts = test_data['test_xvals']

    binw = np.mean(xpts[1:] - xpts[0:-1])
    atol = kwargs.get('atol_diff', 1e-2)/binw
    atol2 = kwargs.get('atol_diff2', 1e-2)/binw

    ens1 = ens_orig.convert_to(gen_class, **test_data['convert_data'])
    diffs = ens_orig.pdf(xpts) - ens1.pdf(xpts)
    assert_all_small(diffs, atol=atol, test_name="convert")

    ens2 = convert(ens_orig, gen_class.name, **test_data['convert_data'])
    diffs = ens_orig.pdf(xpts) - ens2.pdf(xpts)
    assert_all_small(diffs, atol=atol2, test_name="convert2")

    assert ens_orig.shape[:-1] == ens1.shape[:-1]
    assert ens_orig.shape[:-1] == ens2.shape[:-1]
    assert ens_orig.frozen.shape[:-1] == ens2.shape[:-1]
    if hasattr(ens2.dist, 'shape'):
        assert ens2.dist.shape[:-1] == ens2.shape[:-1]


def plotting_func_tests(ensemble, do_samples=False):
    """Run the test for a practicular class"""
    pdf = ensemble[0]
    fig, axes = plot_native(pdf, xlim=(-5, 5))
    assert fig is not None
    assert axes is not None
    fig, axes = plot(pdf, axes=axes)
    assert fig is not None
    assert axes is not None
    fig, axes = plot_native(pdf.frozen,  xlim=(-5, 5))
    assert fig is not None
    assert axes is not None
    fig, axes = plot(pdf.frozen,  xlim=(-5, 5))
    assert fig is not None
    assert axes is not None
    if do_samples:
        samples = pdf.rvs(size=1000)
        plot_pdf_samples_on_axes(axes, pdf, samples)


def run_plotting_func_tests(test_data, do_samples=False):
    """Run the test for a practicular class"""
    ens = build_ensemble(test_data)
    if ens.ndim != 1:
        return
    plotting_func_tests(ens, do_samples=do_samples)



"""Integer packing utilities for qp"""

import enum
import numpy as np



class PackingType(enum.Enum):

    linear_from_rowmax = 0
    log_from_rowmax = 1


def linear_pack_from_rowmax(input_array):
    """Pack an array into 8bit unsigned integers, using the maximum of each row as a refrence

    This packs the values onto a linear grid for each row, running from 0 to row_max

    Parameters
    ----------
    input_array : array_like
        The values we are packing

    Returns
    -------
    packed_array : array_like
        The packed values
    row_max : array_like
        The max for each row, need to unpack the array
    """
    row_max = np.expand_dims(input_array.max(axis=1), -1)
    return np.round(255 * input_array / row_max).astype(np.uint8), row_max


def linear_unpack_from_rowmax(packed_array, row_max):
    """Unpack an array into 8bit unsigned integers, using the maximum of each row as a refrence

    Parameters
    ----------
    packed_array : array_like
        The packed values
    row_max : array_like
        The max for each row, need to unpack the array


    Returns
    -------
    unpacked_array : array_like
        The unpacked values
    """
    unpacked_array = row_max * packed_array / 255.
    return unpacked_array


def log_pack_from_rowmax(input_array, log_floor=-3.):
    """Pack an array into 8bit unsigned integers, using the maximum of each row as a refrence

    This packs the values onto a log grid for each row, running from row_max / 10**log_floor to row_max

    Parameters
    ----------
    input_array : array_like
        The values we are packing
    log_floor: float
        The logarithmic floor used for the packing

    Returns
    -------
    packed_array : array_like
        The packed values
    row_max : array_like
        The max for each row, need to unpack the array
    """
    neg_log_floor = -1. * log_floor
    epsilon = np.power(10., 3 * log_floor)
    row_max = np.expand_dims(input_array.max(axis=1), -1)
    return np.round(255*(np.log10((input_array + epsilon) / row_max) + neg_log_floor)/neg_log_floor).clip(0., 255.).astype(np.uint8), row_max


def log_unpack_from_rowmax(packed_array, row_max, log_floor=-3.):
    """Unpack an array into 8bit unsigned integers, using the maximum of each row as a refrence

    Parameters
    ----------
    packed_array : array_like
        The packed values
    row_max : array_like
        The max for each row, need to unpack the array
    log_floor: float
        The logarithmic floor used for the packing

    Returns
    -------
    unpacked_array : array_like
        The unpacked values
    """
    neg_log_floor = -1. * log_floor
    unpacked_array = row_max * np.where(packed_array == 0, 0., np.power(10, neg_log_floor * ( (packed_array / 255.) - 1. ) ))
    return unpacked_array


def pack_array(packing_type, input_array, **kwargs):
    """Pack an array into 8bit unsigned integers

    Parameters
    ----------
    packing_type : PackingType
        Enum specifing the type of packing to use
    input_array : array_like
        The values we are packing

    Return values and keyword argument depend on the packing type used
    """

    if packing_type == PackingType.linear_from_rowmax:
        return linear_pack_from_rowmax(input_array)
    if packing_type == PackingType.log_from_rowmax:
        return log_pack_from_rowmax(input_array, kwargs.get('log_floor', -3))
    raise ValueError(f"Packing for packing type {packing_type} is not implemetned")  # pragma: no cover


def unpack_array(packing_type, packed_array, **kwargs):
    """Unpack an array from 8bit unsigned integers

    Parameters
    ----------
    packing_type : PackingType
        Enum specifing the type of packing to use
    packed_array : array_like
        The packed values

    Return values and keyword argument depend on the packing type used
    """
    if packing_type == PackingType.linear_from_rowmax:
        return linear_unpack_from_rowmax(packed_array, row_max=kwargs.get('row_max'))
    if packing_type == PackingType.log_from_rowmax:
        return log_unpack_from_rowmax(packed_array, row_max=kwargs.get('row_max'), log_floor=kwargs.get('log_floor', -3))
    raise ValueError(f"Unpacking for packing type {packing_type} is not implemetned")  # pragma: no cover



"""This module implements metric calculations that are independent of qp.Ensembles"""

import numpy as np
from scipy import stats
from scipy.integrate import quad
from scipy.optimize import minimize_scalar

from qp.utils import safelog

def quick_anderson_ksamp(p_random_variables, q_random_variables, **kwargs):
    """Calculate the k-sample Anderson-Darling statistic using scipy.stats.anderson_ksamp for two CDFs. 
    For more details see: https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.anderson_ksamp.html

    Parameters
    ----------
    p_random_variables : np.array
        An array of random variables from the given distribution
    q_random_variables : np.array
        An array of random variables from the given distribution

    Returns
    -------
    [Result objects]
        A array of objects with attributes ``statistic``, ``critical_values``, and ``significance_level``.
    """
    return stats.anderson_ksamp([p_random_variables, q_random_variables], **kwargs)

def quick_kld(p_eval, q_eval, dx=0.01):
    """
    Calculates the Kullback-Leibler Divergence between two evaluations of PDFs.

    Parameters
    ----------
    p_eval: numpy.ndarray, float
        evaluations of probability distribution whose distance _from_ `q` will be calculated
    q_eval: numpy.ndarray, float
        evaluations of probability distribution whose distance _to_ `p` will be calculated.
    dx: float
        resolution of integration grid

    Returns
    -------
    Dpq: float
        the value of the Kullback-Leibler Divergence from `q` to `p`
    """

    # safelog would be easy to isolate if array_metrics is ever extracted
    logquotient = safelog(p_eval) - safelog(q_eval)

    # Calculate the KLD from q to p
    Dpq = dx * np.sum(p_eval * logquotient, axis=-1)
    return Dpq

def quick_moment(p_eval, grid_to_N, dx):
    """
    Calculates a moment of an evaluated PDF

    Parameters
    ----------
    p_eval: numpy.ndarray, float
        the values of a probability distribution
    grid: numpy.ndarray, float
        the grid upon which p_eval was evaluated
    dx: float
        the difference between regular grid points
    N: int
        order of the moment to be calculated

    Returns
    -------
    M: float
        value of the moment
    """
    M = np.dot(p_eval, grid_to_N) * dx
    return M

def quick_rmse(p_eval, q_eval, N):
    """
    Calculates the Root Mean Square Error between two evaluations of PDFs.

    Parameters
    ----------
    p_eval: numpy.ndarray, float
        evaluation of probability distribution function whose distance between
        its truth and the approximation of `q` will be calculated.
    q_eval: numpy.ndarray, float
        evaluation of probability distribution function whose distance between
        its approximation and the truth of `p` will be calculated.
    N: int
        number of points at which PDFs were evaluated

    Returns
    -------
    rms: float
        the value of the RMS error between `q` and `p`
    """
    # Calculate the RMS between p and q
    rms = np.sqrt(np.sum((p_eval - q_eval) ** 2, axis=-1) / N)
    return rms

def quick_rbpe(pdf_function, integration_bounds, limits=(np.inf, np.inf)):
    """
    Calculates the risk based point estimate of a qp.Ensemble object with npdf == 1.

    Parameters
    ----------
    pdf_function, python function
        The function should calculate the value of a pdf at a given x value
    integration_bounds, 2-tuple of floats
        The integration bounds - typically (ppf(0.01), ppf(0.99)) for the given distribution
    limits, tuple of floats
        The limits at which to evaluate possible z_best estimates.
        If custom limits are not provided then all potential z value will be
        considered using the scipy.optimize.minimize_scalar function.

    Returns
    -------
    rbpe: float
        The risk based point estimate of the provided ensemble.
    """

    def calculate_loss(x):
        return 1.0 - (1.0 / (1.0 + (pow((x / 0.15), 2))))

    lower = integration_bounds[0]
    upper = integration_bounds[1]

    def find_z_risk(zp):
        def integrand(z):
            return pdf_function(z) * calculate_loss((zp - z) / (1.0 + z))

        return quad(integrand, lower, upper)[0]

    if limits[0] == np.inf:
        return minimize_scalar(find_z_risk).x
    return minimize_scalar(
        find_z_risk, bounds=(limits[0], limits[1]), method="bounded"
    ).x



import logging
import numpy as np
from scipy import stats
import qp
from qp.metrics.metrics import calculate_outlier_rate
from qp.metrics.array_metrics import quick_anderson_ksamp

DEFAULT_QUANTS = np.linspace(0, 1, 100)

class PIT():
    """PIT(qp_ens, true_vals, eval_grid=DEFAULT_QUANTS)
    Probability Integral Transform

    Parameters
    ----------
    qp_ens : Ensemble
        A collection of N distribution objects
    true_vals : [float]
        An array-like sequence of N float values representing the known true value for each distribution
    eval_grid : [float], optional
        A strictly increasing array-like sequence in the range [0,1], by default DEFAULT_QUANTS

    Returns
    -------
    PIT object
        An object with an Ensemble containing the PIT distribution, and a full set of PIT samples.
    """

    def __init__(self, qp_ens, true_vals, eval_grid=DEFAULT_QUANTS):
        """We will create a quantile Ensemble to store the PIT distribution, but also store the
        full set of PIT samples as ancillary data of the (single PDF) ensemble.

        Parameters
        ----------
        qp_ens : Ensemble
            A collection of N distribution objects
        true_vals : [float]
            An array-like sequence of N float values representing the known true value for each distribution
        eval_grid : [float], optional
            A strictly increasing array-like sequence in the range [0,1], by default DEFAULT_QUANTS
        """

        self._true_vals = true_vals

        # For each distribution in the Ensemble, calculate the CDF where x = known_true_value
        self._pit_samps = np.array([qp_ens[i].cdf(self._true_vals[i])[0][0] for i in range(len(self._true_vals))])

        n_pit = np.min([len(self._pit_samps), len(eval_grid)])
        if n_pit < len(eval_grid):
            logging.warning('Number of pit samples is smaller than the evaluation grid size. '
                            'Will create a new evaluation grid with size = number of pit samples')
            eval_grid = np.linspace(0, 1, n_pit)

        data_quants = np.quantile(self._pit_samps, eval_grid)
        self._pit = qp.Ensemble(qp.quant, data=dict(quants=eval_grid, locs=np.atleast_2d(data_quants)))

    @property
    def pit_samps(self):
        """Returns the PIT samples. i.e. ``CDF(true_vals)`` for each distribution in the Ensemble used to initialize the PIT object.

        Returns
        -------
        np.array
            An array of floats
        """
        return self._pit_samps

    @property
    def pit(self):
        """Return the PIT Ensemble object

        Returns
        -------
        qp.Ensemble
            An Ensemble containing 1 qp.quant distribution.
        """
        return self._pit

    def calculate_pit_meta_metrics(self):
        """Convenience method that will calculate all of the PIT meta metrics and return them 
            as a dictionary.

        Returns
        -------
        dictionary
            The collection of PIT statistics
        """
        pit_meta_metrics = {}

        pit_meta_metrics['ad'] = self.evaluate_PIT_anderson_ksamp()
        pit_meta_metrics['cvm'] = self.evaluate_PIT_CvM()
        pit_meta_metrics['ks'] = self.evaluate_PIT_KS()
        pit_meta_metrics['outlier_rate'] = self.evaluate_PIT_outlier_rate()

        return pit_meta_metrics

    def evaluate_PIT_anderson_ksamp(self, pit_min=0., pit_max=1.):
        """Use scipy.stats.anderson_ksamp to compute the Anderson-Darling statistic
        for the cdf(truth) values by comparing with a uniform distribution between 0 and 1.
        Up to the current version (1.9.3), scipy.stats.anderson does not support
        uniform distributions as reference for 1-sample test, therefore we create a uniform
        "distribution" and pass it as the second value in the list of parameters to the scipy
        implementation of k-sample Anderson-Darling.
        For details see: https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.anderson_ksamp.html

        Parameters
        ----------
        pit_min : float, optional
            Minimum PIT value to accept, by default 0.
        pit_max : float, optional
            Maximum PIT value to accept, by default 1.

        Returns
        -------
        [Result objects]
            A array of objects with attributes `statistic`, `critical_values`, and `significance_level`.
        For details see: https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.anderson_ksamp.html
        """
        # Removed the CDF values that are outside the min/max range
        trimmed_pit_values = self._trim_pit_values(pit_min, pit_max)

        uniform_yvals = np.linspace(pit_min, pit_max, len(trimmed_pit_values))

        return quick_anderson_ksamp(trimmed_pit_values, uniform_yvals)

    def evaluate_PIT_CvM(self):
        """Calculate the Cramer von Mises statistic using scipy.stats.cramervonmises using self._pit_samps
        compared to a uniform distribution. For more details see:
        https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.cramervonmises.html

        Returns
        -------
        [Result objects]
            A array of objects with attributes `statistic` and `pvalue`
            For details see: https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.cramervonmises.html
        """
        return stats.cramervonmises(self._pit_samps, stats.uniform.cdf)

    def evaluate_PIT_KS(self):
        """Calculate the Kolmogorov-Smirnov statistic using scipy.stats.kstest. For more details see:
        https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.kstest.html

        Returns
        -------
        [Return Object]
            A array of objects with attributes `statistic` and `pvalue`.
            For details see: https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.kstest.html
        """
        return stats.kstest(self._pit_samps, stats.uniform.cdf)

    def evaluate_PIT_outlier_rate(self, pit_min=0.0001, pit_max=0.9999):
        """Compute fraction of PIT outliers by evaluating the CDF of the distribution in the PIT Ensemble
        at `pit_min` and `pit_max`.

        Parameters
        ----------
        pit_min : float, optional
            Lower bound for outliers, by default 0.0001
        pit_max : float, optional
            Upper bound for outliers, by default 0.9999

        Returns
        -------
        float
            The percentage of outliers in this distribution given the min and max bounds.
        """
        return calculate_outlier_rate(self._pit, pit_min, pit_max)[0]

    def _trim_pit_values(self, cdf_min, cdf_max):
        """Remove and report any cdf(x) that are outside the min/max range.

        Parameters
        ----------
        cdf_min : float
            The minimum cdf(x) value to accept
        cdf_max : float
            The maximum cdf(x) value to accept

        Returns
        -------
        [float]
            The list of PIT values within the min/max range.
        """
        # Create truth mask for pit values between cdf_min and pit max
        mask = (self._pit_samps >= cdf_min) & (self._pit_samps <= cdf_max)

        # Keep pit values that are within the min/max range
        pits_clean = self._pit_samps[mask]

        # Determine how many pit values were dropped and warn the user.
        diff = len(self._pit_samps) - len(pits_clean)
        if diff > 0:
            logging.warning("Removed %d PITs from the sample.", diff)

        return pits_clean



from .array_metrics import *
from .metrics import *
from .goodness_of_fit import *

# added for testing purposes
from .metrics import _calculate_grid_parameters, _check_ensemble_is_not_nested, _check_ensembles_are_same_size
from .metrics import _check_ensembles_contain_correct_number_of_distributions
